﻿/******************************************************************************
Module:  MyPage.xaml.cs
Notices: Copyright (c) by Jeffrey Richter and Wintellect
******************************************************************************/

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Threading.Tasks;
using Windows.ApplicationModel.Activation;
using Windows.ApplicationModel.Contacts;
using Windows.ApplicationModel.Core;
using Windows.ApplicationModel.DataTransfer;
using Windows.Data.Xml.Dom;
using Windows.Foundation.Collections;
using Windows.Storage;
using Windows.Storage.AccessCache;
using Windows.Storage.Pickers;
using Windows.Storage.Provider;
using Windows.System;
using Windows.UI.Core;
using Windows.UI.Notifications;
using Windows.UI.Popups;
using Windows.UI.ViewManagement;
using Windows.UI.Xaml;
using Windows.UI.Xaml.Controls;
using Windows.UI.Xaml.Navigation;
using Wintellect.Settings;
using Wintellect.WinRT.AppAids;
using Wintellect.WinRT.FramePageStateManager;
using Wintellect.WinRT.Xaml;

public sealed class SuspendSettings : INotifyPropertyChanged {
   private readonly ISettingStore<SuspendSettings> m_store;
   public SuspendSettings(ISettingStore<SuspendSettings> settingStore) { m_store = settingStore; }

   public Int32 PageNumber {
      get { return m_store.Get(0); }
      set { m_store.Set(value); }
   }

   public String ValueEnteredOnPage {
      get { return m_store.Get(String.Empty); }
      set { m_store.Set(value); }
   }

   public void RemoveSettings() { m_store.RemoveSettings(); }

   public event PropertyChangedEventHandler PropertyChanged {
      add { m_store.PropertyChanged += value; }
      remove { m_store.PropertyChanged -= value; }
   }
}

namespace ProcessModel {
   public sealed partial class MyPage : Page {
      private SuspendSettings Settings;

      public MyPage() {
         InitializeComponent();
#if WINDOWS_PHONE_APP
         m_btnNewView.Visibility = Visibility.Collapsed;
         m_btnContactPicker.Visibility = Visibility.Collapsed;
#endif
      }

      protected override void OnNavigatedTo(NavigationEventArgs e) {
         m_txtArgumentPassedToPage.Text = e.Parameter.ToString(); // Show passed-in arg

         // Get this page's session state dictionary
         var pageState = this.GetCurrentPageState(e.NavigationMode == NavigationMode.New);

         // Create a strongly-typed Settings object backed by the dictionary store
         Settings = new SuspendSettings(new DictionarySettingStore<SuspendSettings>(pageState));

         // Save some state
         Settings.PageNumber = Frame.BackStackDepth;
         m_txtPageNum.Text = "Page #" + Settings.PageNumber;

         // Data bind the Text control to Setting's SomeText so it automatically stays in sync
         m_txtValueEnteredOnPage.DataContext = Settings;
         m_txtPrevExecutionState.DataContext = this;
         base.OnNavigatedTo(e);
      }

      protected override void OnNavigatedFrom(NavigationEventArgs e) {
         base.OnNavigatedFrom(e);
      }

      private void Crash(object sender, RoutedEventArgs e) {
         throw new InvalidOperationException("Purposeful app crash");
      }

      private void GoBack(object sender, RoutedEventArgs e) { Frame.GoBack(); }
      private void GoForward(object sender, RoutedEventArgs e) { Frame.GoForward(); }
      private void NewPage(object sender, RoutedEventArgs e) { Frame.Navigate(typeof(MyPage), Settings.ValueEnteredOnPage); }

      #region Activation Tests
      public String PreviousExecutionState { get { return App.PreviousExecutionState.ToString(); } }

      private async void FileActivate(object sender, RoutedEventArgs e) {
         StorageFile sf = await ApplicationData.Current.TemporaryFolder.CreateFileAsync("temp.wintellect", CreationCollisionOption.ReplaceExisting);
            //StorageFile.CreateStreamedFileAsync("Whatever.wintellect", sfdr => { sfdr.Dispose(); }, null);
         Boolean t = await Launcher.LaunchFileAsync(sf);
      }
      private async void Protocol(object sender, RoutedEventArgs e) {
         Boolean t = await Launcher.LaunchUriAsync(new Uri("wintellect://foo"));
      }



      private async void FileOpen(object sender, RoutedEventArgs e) {
         var picker = new FileOpenPicker { FileTypeFilter = { ".txt" } };
#if WINDOWS_APP
         var file = await picker.PickSingleFileAsync();
#elif WINDOWS_PHONE_APP
         picker.PickSingleFileAndContinue();
         await Task.FromResult(true);  // Remove compiler warning
#endif
      }


      private async void FileSave(object sender, RoutedEventArgs e) {
         var picker = new FileSavePicker {
            FileTypeChoices = { { "Wintellect files", new[] { ".wintellect" } } }
         };
#if WINDOWS_APP
         await picker.PickSaveFileAsync();
#elif WINDOWS_PHONE_APP
         picker.PickSaveFileAndContinue();
         await Task.FromResult(true);  // Remove compiler warning
#endif
      }


      private void Share(object sender, RoutedEventArgs e) {
         DataTransferManager.ShowShareUI();
      }
      private async void Contact(object sender, RoutedEventArgs e) {
         Contact contact = await new ContactPicker().PickContactAsync();
      }

      private void Toast(object sender, RoutedEventArgs e) {
         if (ToastNotificationManager.CreateToastNotifier().Setting != NotificationSetting.Enabled) {
            new MessageDialog("You must enable toast notifications for this app").ShowAsync().Forget();
         } else {
            XmlDocument xml = new XmlDocument();
            xml.LoadXml("<toast launch=\"???\"><visual><binding template=\"ToastText01\"><text id=\"1\">Click Me!</text></binding></visual></toast>");
            var tn = new ToastNotification(xml);
            ToastNotificationManager.CreateToastNotifier().Show(tn);
         }
      }

      private IActivatedEventArgs m_args = null;
      private async void AddCachedFile(object sender, RoutedEventArgs e) {
         var args = (FileOpenPickerActivatedEventArgs)m_args;
         var sf = await ApplicationData.Current.LocalFolder.CreateFileAsync("CachedFile.plm", CreationCollisionOption.ReplaceExisting);
         await FileIO.WriteTextAsync(sf, "Cached file created...");
         CachedFileUpdater.SetUpdateInformation(sf, sf.DisplayName, ReadActivationMode.BeforeAccess,
            WriteActivationMode.AfterWrite, CachedFileOptions.RequireUpdateOnAccess);
         var token = StorageApplicationPermissions.FutureAccessList.Add(sf);
         sf = await StorageApplicationPermissions.FutureAccessList.GetFileAsync(token);
         var text = await FileIO.ReadTextAsync(sf);
         var dd = args.FileOpenPickerUI.AddFile(sf.Name, sf);
      }
      #endregion

#if WINDOWS_PHONE_APP
      private void NewView(object sender, RoutedEventArgs e) { }
#elif WINDOWS_APP
      private async void NewView(object sender, RoutedEventArgs e) {
         // Have Windows create a new CoreApplicationView (thread, CoreWindow, & CoreDispatcher)
         CoreApplicationView cav = CoreApplication.CreateNewView();

         CoreWindow newAppViewWindow = null; // This will hold the new view's window

         // Have the new thread initialize the new view's content
         await cav.Dispatcher.RunAsync(CoreDispatcherPriority.Normal, () => {
            // Give the new thread's window back to the creating thread
            newAppViewWindow = Window.Current.CoreWindow;

            // Create the desired UI element tree & make it the content of the new window
            Window.Current.Content = new MyPage();
            Window.Current.Activate();

            // Give the ID of the new thread's view back to the creating thread
            ApplicationView newAppView = ApplicationView.GetForCurrentView();
            newAppView.Title = "View #" + newAppView.Id;
            newAppView.IsScreenCaptureEnabled = false;
            newAppView.Consolidated += (appView, args) => { var x = args.IsUserInitiated; };
         });

         // After the new thread initializes its view, the creating thread makes it appear
         Int32 newAppViewId = ApplicationView.GetApplicationViewIdForWindow(newAppViewWindow);
         await ApplicationViewSwitcher.TryShowAsStandaloneAsync(newAppViewId, ViewSizePreference.UseLess);
      }
#endif
   }

   public sealed class BooleanVisibilityConverter : ConverterBase {
      public override Object Convert(Object value, Type targetType, Object parameter, String culture) {
         if (value == null) return DependencyProperty.UnsetValue;
         return ((Boolean)value) ? Visibility.Collapsed : Visibility.Visible;
      }
   }
}
