﻿/******************************************************************************
Module:  ContinuationManager.cs
Notices: Copyright (c) by Jeffrey Richter and Wintellect
******************************************************************************/

// Install "Shared Project Reference Manager" from http://visualstudiogallery.msdn.microsoft.com/315c13a7-2787-4f57-bdf7-adae6ed54450 
// to add this shared project to other projects

using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using System.Reflection;
using System.Runtime.CompilerServices;
using System.Threading;
using System.Threading.Tasks;
using Windows.ApplicationModel.Activation;
using Windows.Foundation;
using Windows.Foundation.Collections;
using Windows.Foundation.Metadata;
using Windows.Security.Authentication.Web;
using Windows.Storage;
using Windows.Storage.Pickers;

#if WINDOWS_APP
using ValueSet = Windows.Foundation.Collections.PropertySet;
namespace Windows.ApplicationModel.Activation {
   public interface IContinuationActivatedEventArgs : IActivatedEventArgs {
      ValueSet ContinuationData { get; }
   }

   public interface IFileOpenPickerContinuationEventArgs : IContinuationActivatedEventArgs, IActivatedEventArgs {
      IReadOnlyList<StorageFile> Files { get; }
   }

   public sealed class FileOpenPickerContinuationEventArgs : IFileOpenPickerContinuationEventArgs, IContinuationActivatedEventArgs, IActivatedEventArgs {
      internal FileOpenPickerContinuationEventArgs(ValueSet continuationData, IReadOnlyList<StorageFile> files) {
         ContinuationData = continuationData;
         Files = files;
      }
      public ValueSet ContinuationData { get; private set; }
      public IReadOnlyList<StorageFile> Files { get; private set; }
      public ActivationKind Kind { get { return (ActivationKind)1002; } }
      public ApplicationExecutionState PreviousExecutionState { get { return ApplicationExecutionState.Running; } }
      public SplashScreen SplashScreen { get { throw new NotImplementedException(); } }
   }

   public interface IFileSavePickerContinuationEventArgs : IContinuationActivatedEventArgs, IActivatedEventArgs {
      StorageFile File { get; }
   }

   public sealed class FileSavePickerContinuationEventArgs : IFileSavePickerContinuationEventArgs, IContinuationActivatedEventArgs, IActivatedEventArgs {
      internal FileSavePickerContinuationEventArgs(ValueSet continuationData, StorageFile file) {
         ContinuationData = continuationData;
         File = file;
      }
      public ValueSet ContinuationData { get; private set; }
      public StorageFile File { get; private set; }
      public ActivationKind Kind { get { return (ActivationKind)1003; } }
      public ApplicationExecutionState PreviousExecutionState { get { return ApplicationExecutionState.Running; } }
      public SplashScreen SplashScreen { get { throw new InvalidOperationException(); } }
   }

   public interface IFolderPickerContinuationEventArgs : IContinuationActivatedEventArgs, IActivatedEventArgs {
      StorageFolder Folder { get; }
   }

   public sealed class FolderPickerContinuationEventArgs : IFolderPickerContinuationEventArgs, IContinuationActivatedEventArgs, IActivatedEventArgs {
      internal FolderPickerContinuationEventArgs(ValueSet continuationData, StorageFolder folder) {
         ContinuationData = continuationData;
         Folder = folder;
      }
      public ValueSet ContinuationData { get; private set; }
      public StorageFolder Folder { get; private set; }
      public ActivationKind Kind { get { return (ActivationKind)1004; } }
      public ApplicationExecutionState PreviousExecutionState { get { return ApplicationExecutionState.Running; } }
      public SplashScreen SplashScreen { get { throw new InvalidOperationException(); } }
   }

   public interface IWebAuthenticationBrokerContinuationEventArgs : IContinuationActivatedEventArgs, IActivatedEventArgs {
      WebAuthenticationResult WebAuthenticationResult { get; }
   }

   public sealed class WebAuthenticationBrokerContinuationEventArgs : IWebAuthenticationBrokerContinuationEventArgs, IContinuationActivatedEventArgs, IActivatedEventArgs {
      internal WebAuthenticationBrokerContinuationEventArgs(ValueSet continuationData, WebAuthenticationResult result) {
         ContinuationData = continuationData;
         WebAuthenticationResult = result;
      }
      public ValueSet ContinuationData { get; private set; }
      public ActivationKind Kind { get { return (ActivationKind)1005; } }
      public ApplicationExecutionState PreviousExecutionState { get { return ApplicationExecutionState.Running; } }
      public SplashScreen SplashScreen { get { throw new InvalidOperationException(); } }
      public WebAuthenticationResult WebAuthenticationResult { get; private set; }
   }
}
#endif

namespace Windows.Security.Authentication.Web {
   public sealed class WebAuthBroker {
      private readonly ValueSet m_continuationData = new ValueSet();
      public WebAuthBroker() { }
      public ValueSet ContinuationData { get { return m_continuationData; } }
      public void AuthenticateAndContinue(Uri requestUri, Uri callbackUri,
         WebAuthenticationOptions options,
         Action<WebAuthenticationBrokerContinuationEventArgs> callback) {
#if WINDOWS_APP
      var ts = TaskScheduler.FromCurrentSynchronizationContext();
      WebAuthenticationBroker.AuthenticateAsync(options, requestUri, callbackUri).AsTask().ContinueWith(
         t => callback(new WebAuthenticationBrokerContinuationEventArgs(m_continuationData, t.Result)),
         CancellationToken.None, TaskContinuationOptions.None, ts);
#elif WINDOWS_PHONE_APP
         ContinuationData.Add("@ContinuationCallbackKey@", callback.GetMethodInfo().Name);
         WebAuthenticationBroker.AuthenticateAndContinue(requestUri, callbackUri, m_continuationData, options);
#endif
      }

#if WINDOWS_PHONE_APP
      public IAsyncOperation<WebAuthenticationResult> AuthenticateSilentlyAsync(Uri requestUri, WebAuthenticationOptions options = WebAuthenticationOptions.None) {
         return Windows.Security.Authentication.Web.WebAuthenticationBroker.AuthenticateSilentlyAsync(requestUri, options);
      }
#endif
      public static Uri GetCurrentApplicationCallbackUri() {
         return Windows.Security.Authentication.Web.WebAuthenticationBroker.GetCurrentApplicationCallbackUri();
      }
   }
   public static class WebAuthExtensions {
      public static IDictionary<String, String> ResultToClaims(this WebAuthenticationResult result) {
         String[] claimNamesAndValues = result.ResponseData.Substring(result.ResponseData.IndexOf('#') + 1).Split('=', '&');
         var claims = new Dictionary<String, String>();
         for (Int32 claim = 0; claim < claimNamesAndValues.Length; claim += 2) {
            claims.Add(claimNamesAndValues[claim], claimNamesAndValues[claim + 1]);
         }
         return claims;
      }
   }
}

namespace Wintellect.WinRT.AppAids {
   public static class ContinuationManager {
#if WINDOWS_PHONE_APP
      private const String c_ContinuationCallbackKey = "@ContinuationCallbackKey@";
      private static void SetContinuation<T>(ValueSet continuationData, Action<T> callback) {
         continuationData.Add(c_ContinuationCallbackKey, callback.GetMethodInfo().Name);
      }
#endif

      #region ContinuationData Emulation
#if WINDOWS_APP
      private readonly static ConditionalWeakTable<Object, ValueSet> s_ContinuationData =
         new ConditionalWeakTable<Object, ValueSet>();
      private static ValueSet GetContinuationDataInternal(this Object picker) {
         ValueSet continuationData;
         if (!s_ContinuationData.TryGetValue(picker, out continuationData)) {
            s_ContinuationData.Add(picker, continuationData = new ValueSet());
         }
         return continuationData;
      }
#endif

      public static ValueSet GetContinuationData(this FileOpenPicker picker) {
#if WINDOWS_APP
         return picker.GetContinuationDataInternal();
#elif WINDOWS_PHONE_APP
         return picker.ContinuationData;
#endif
      }
      public static ValueSet GetContinuationData(this FileSavePicker picker) {
#if WINDOWS_APP
         return picker.GetContinuationDataInternal();
#elif WINDOWS_PHONE_APP
         return picker.ContinuationData;
#endif
      }
      public static ValueSet GetContinuationData(this FolderPicker picker) {
#if WINDOWS_APP
         return picker.GetContinuationDataInternal();
#elif WINDOWS_PHONE_APP
         return picker.ContinuationData;
#endif
      }
      #endregion

      public static void PickSingleFileAndContinue(this FileOpenPicker picker, Action<FileOpenPickerContinuationEventArgs> callback) {
#if WINDOWS_APP
         var ts = TaskScheduler.FromCurrentSynchronizationContext();
         picker.PickSingleFileAsync().AsTask().ContinueWith(
            t => callback(new FileOpenPickerContinuationEventArgs(picker.GetContinuationData(), new List<StorageFile> { t.Result })),
            CancellationToken.None, TaskContinuationOptions.None, ts);
#elif WINDOWS_PHONE_APP
         SetContinuation(picker.ContinuationData, callback);
         picker.PickSingleFileAndContinue();
#endif
      }

      public static void PickMultipleFilesAndContinue(this FileOpenPicker picker, Action<FileOpenPickerContinuationEventArgs> callback) {
#if WINDOWS_APP
         var ts = TaskScheduler.FromCurrentSynchronizationContext();
         picker.PickMultipleFilesAsync().AsTask().ContinueWith(
            t => callback(new FileOpenPickerContinuationEventArgs(picker.GetContinuationData(), t.Result)),
            CancellationToken.None, TaskContinuationOptions.None, ts);
#elif WINDOWS_PHONE_APP
         SetContinuation(picker.ContinuationData, callback);
         picker.PickMultipleFilesAndContinue();
#endif
      }

      public static void PickSaveFileAndContinue(this FileSavePicker picker, Action<FileSavePickerContinuationEventArgs> callback) {
#if WINDOWS_APP
         var ts = TaskScheduler.FromCurrentSynchronizationContext();
         picker.PickSaveFileAsync().AsTask().ContinueWith(
            t => callback(new FileSavePickerContinuationEventArgs(picker.GetContinuationData(), t.Result)),
            CancellationToken.None, TaskContinuationOptions.None, ts);
#elif WINDOWS_PHONE_APP
         SetContinuation(picker.ContinuationData, callback);
         picker.PickSaveFileAndContinue();
#endif
      }

      public static void PickFolderAndContinue(this FolderPicker picker, Action<FolderPickerContinuationEventArgs> callback) {
#if WINDOWS_APP
         var ts = TaskScheduler.FromCurrentSynchronizationContext();
         picker.PickSingleFolderAsync().AsTask().ContinueWith(
            t => callback(new FolderPickerContinuationEventArgs(picker.GetContinuationData(), t.Result)),
            CancellationToken.None, TaskContinuationOptions.None, ts);
#elif WINDOWS_PHONE_APP
         SetContinuation(picker.ContinuationData, callback);
         picker.PickFolderAndContinue();
#endif
      }

      public static Boolean OnActivatedForContinuation(this IActivatedEventArgs args, Object target) {
#if WINDOWS_APP
         return false;
#elif WINDOWS_PHONE_APP
         IContinuationActivatedEventArgs cea = args as IContinuationActivatedEventArgs;
         if (cea == null) return false;

         Object methodName;
         if (!cea.ContinuationData.TryGetValue(c_ContinuationCallbackKey, out methodName))
            throw new InvalidOperationException("ContinuationData is missing the " + c_ContinuationCallbackKey + " key.");
         Type argType = null;
         switch (args.Kind) {
            case ActivationKind.PickFileContinuation: argType = typeof(FileOpenPickerContinuationEventArgs); break;
            case ActivationKind.PickSaveFileContinuation: argType = typeof(FileSavePickerContinuationEventArgs); break;
            case ActivationKind.PickFolderContinuation: argType = typeof(FolderPickerContinuationEventArgs); break;
            case ActivationKind.WebAuthenticationBrokerContinuation: argType = typeof(WebAuthenticationBrokerContinuationEventArgs); break;
         }
         // cea.ContinuationData.Remove(c_ContinuationCallbackKey);  // WinRT throws UnauthorizedException when trying to do this
         var mi = target.GetType().GetTypeInfo().GetDeclaredMethods((String)methodName).FirstOrDefault(m => {
            ParameterInfo[] parameters = m.GetParameters();
            return (parameters.Length == 1) && (parameters[0].ParameterType == argType);
         });
         if (mi == null) throw new InvalidOperationException(String.Format("The {0} type doesn't define a {1} method.", target.GetType(), methodName));
         mi.Invoke(target, new Object[] { args });
         return true;
#endif
      }
   }
}